"use strict";

export default class Widget {

    /** @member {HTMLElement} */
    _element

    /** @member {Object} */
    _options = {};

    /** @member {WidgetManager} */
    _widgetManager

    /** @member {String} */
     _name

     /**
     * @param {HTMLElement} element
     * @param {String} name
     * @param {WidgetManager} widgetManager
     * @param {Object|null=} listeners
     * @param {Object|null=} options
     */
    constructor(
        element,
        name,
        widgetManager,
        listeners,
        options
    ) {
        this._element = element;
        this._widgetManager = widgetManager;
        this._name = name;

        if (listeners) {
            this.#attachEventListeners(listeners);
        }

        if (!options) {
            options = {};
        }

        if (options) {
            this._options = {...options, ...this.#getOptionsFromElement()};
        }
        this._doConstruct();
    }

    /**
     * @returns {void}
     */
    init() {
        if (this._isInitialized()) {
            console.warn('Widget [' + this._name + '] is already initialized!')
            return;
        }
        this._doInit();
        this._element.grifix_widget = this;
        this._widgetManager.bindWidgets(this._element);
        this._dispatchEvent('init');
    }

    /**
     * @returns {String}
     */
    getClassName() {
        return this.constructor.name;
    }

    /**
     *
     * @returns {void}
     * @protected
     */
    _doInit() {

    }

    /**
     *
     * @protected
     * @returns {void}
     */
    _doConstruct() {

    }

    /**
     * @returns {Object}
     */
    #getOptionsFromElement() {
        if (this._element.hasAttribute('data-options')) {
            return JSON.parse(this._element.getAttribute('data-options'));
        }
        return {};
    }

    /**
     * @returns {boolean}
     * @protected
     */
    _isInitialized() {
        if (this._element['grifix_widget']) {
            return true;
        }
        return false;
    }

    /**
     * @param {String=} sourceUrl
     * @returns {Promise<void>}
     */
    async reload(sourceUrl) {

        if (!sourceUrl) {
            sourceUrl = this._getSourceUrl();
        }
        const div = document.createElement('div');
        div.innerHTML = await this.#fetchSource(sourceUrl);
        this._element.innerHTML = div.querySelector(':first-child').innerHTML;
        this._element.grifix_widget = null;
        this.init();
    }

    /**
     * @param listeners
     * @returns {Object}
     */
    #attachEventListeners(listeners) {
        //TODO implement it
    }

    /**
     *
     * @return {String}
     * @protected
     */
    _getSourceUrl() {
        return this._element.getAttribute('data-source_url');
    }

    /**
     * @param {String} eventType
     * @returns {void}
     * @protected
     */
    _dispatchEvent(eventType) {
        this._element.dispatchEvent(new Event(eventType));
    }

    /**
     * @param {String|null} sourceUrl
     * @returns {Promise<string>}
     */
    async #fetchSource(sourceUrl) {
        if (!sourceUrl) {
            sourceUrl = this._getSourceUrl();
        }
        if (!sourceUrl) {
            throw new Error('Source url is not defined for widget [' + this._name + ']!');
        }
        const response = await fetch(sourceUrl);
        if (response.ok) {
            return response.text()
        }
        throw new Error('Cannot fetch source!')
    }


    /**
     *
     * @param {String} name
     * @protected
     * @returns HTMLElement
     */
    _getElement(name) {
        const result = this._element.querySelector('[data-role="' + name + '"]');
        if (!result) {
            throw new Error('Element [' + name + '] does not exist for widget [' + this._name + ']!');
        }
        return result;
    }


    /**
     * @param name
     * @returns {NodeListOf<HTMLElement>}
     * @protected
     */
    _getElements(name) {
        return this._element.querySelectorAll('[data-role="' + name + '"]');
    }
}
